@@ -6,25 +6,37 @@
 import datetime
 import platform
 import scrapy
-from selenium import webdriver
+
+# Due to the problems with selenium wire on linux systems
+try:
+    from selenium import webdriver
+except:
+    print('Error while importing selenium-wire, using selenium instead')
+    from selenium import webdriver
+
 from selenium.webdriver.common.by import By
+from selenium.webdriver.common.proxy import Proxy, ProxyType
 from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
-from selenium.webdriver import FirefoxOptions
+from selenium.webdriver import FirefoxOptions, ChromeOptions
 from selenium.webdriver.support.ui import WebDriverWait
 from selenium.webdriver.support import expected_conditions as EC
 from selenium.webdriver.firefox.service import Service
 from webdriver_manager.firefox import GeckoDriverManager
+from webdriver_manager.chrome import ChromeDriverManager
+import urllib.request
+import requests
 import time
-import pynput.keyboard
+#import pynput.keyboard
 import pickle
 import os
 import shutil
 import re
+import psutil
 
 import lxml.html
 
-keyboard = pynput.keyboard.Controller()
-key = pynput.keyboard.Key
+#keyboard = pynput.keyboard.Controller()
+#key = pynput.keyboard.Key
 
 from scrapy.selector import Selector
 from keepvariable.keepvariable_core import VarSafe, kept_variables, save_variables, load_variable_safe
@@ -61,7 +73,7 @@ def print_special(inp):
     save_variables({"inp": inp}, filename="input.kpv")
 
 
-def take_screenshot(browser, inp):
+def take_screenshot(browser, page, inp):
     """
     Takes screenshot of current page and saves it.
         :param browser: Selenium driver, browser instance
@@ -78,32 +90,104 @@ def take_screenshot(browser, inp):
     except Exception as e:
         print('Error while taking page screenshot!', e)
     '''
-    try:
-        root_element = browser.find_element(By.XPATH, '/html')
-        string = browser.get_full_page_screenshot_as_base64()
-        browser.execute_script("return arguments[0].scrollIntoView(true);", root_element)
+    if type(browser) == webdriver.Firefox:
 
+        try:
+            root_element = browser.find_element(By.XPATH, '/html')
+            string = browser.get_full_page_screenshot_as_base64()
+            browser.execute_script("return arguments[0].scrollIntoView(true);", root_element)
+
+            with open(filename, "w+") as fh:
+                fh.write(string)
+        except Exception as e:
+            print('Error while taking page screenshot!', e)
+
+    elif type(browser) == webdriver.Chrome:
+        # Get params needed for fullpage screenshot
+        page_rect = browser.execute_cdp_cmd('Page.getLayoutMetrics', {})
+
+        # Set the width and height of the viewport to screenshot, same as the site's content size
+        screenshot_config = {'captureBeyondViewport': True,
+                             'fromSurface': True,
+                             'clip': {'width': page_rect['cssContentSize']['width'],
+                                      'height': page_rect['cssContentSize']['height'],
+                                      'x': 0,
+                                      'y': 0,
+                                      'scale': 1},
+                             }
+        # Dictionary with 1 key: data
+        string = browser.execute_cdp_cmd('Page.captureScreenshot', screenshot_config)['data']  # Taking screenshot
         with open(filename, "w+") as fh:
             fh.write(string)
+
+def take_png_screenshot(browser, page, inp):
+    """
+    Takes screenshot of current page and saves it.
+        :param browser: Selenium driver, browser instance
+        :param inp, list, inputs from launcher (filename)
+    """
+
+    filename = inp[0]
+    '''
+    try:
+        root_element = browser.find_element(By.XPATH, '/html')
+        browser.save_full_page_screenshot(filename)
+
+        browser.execute_script("return arguments[0].scrollIntoView(true);", root_element)
     except Exception as e:
         print('Error while taking page screenshot!', e)
+    '''
+    if type(browser) == webdriver.Firefox:
+
+        try:
+            root_element = browser.find_element(By.XPATH, '/html')
+            #string = browser.get_full_page_screenshot_as_file(filename)
+            #driver.find_element_by_tag_name('body').screenshot('web_screenshot4.png')
+    
+            screenshot=browser.get_full_page_screenshot_as_file(filename)
+            print(screenshot)
+            browser.execute_script("return arguments[0].scrollIntoView(true);", root_element)
+
+            #with open(filename, "w+") as fh:
+            #    fh.write(string)
+        except Exception as e:
+            print('Error while taking page screenshot!', e)
+
+    # elif type(browser) == webdriver.Chrome:
+    #     # Get params needed for fullpage screenshot
+    #     page_rect = browser.execute_cdp_cmd('Page.getLayoutMetrics', {})
+
+    #     # Set the width and height of the viewport to screenshot, same as the site's content size
+    #     screenshot_config = {'captureBeyondViewport': True,
+    #                          'fromSurface': True,
+    #                          'clip': {'width': page_rect['cssContentSize']['width'],
+    #                                   'height': page_rect['cssContentSize']['height'],
+    #                                   'x': 0,
+    #                                   'y': 0,
+    #                                   'scale': 1},
+    #                          }
+    #     # Dictionary with 1 key: data
+    #     string = browser.execute_cdp_cmd('Page.captureScreenshot', screenshot_config)['data']  # Taking screenshot
+    #     with open(filename, "w+") as fh:
+    #         fh.write(string)
 
 
-def extract_page_source(page, inp):
+
+
+def extract_page_source(browser, page, inp):
     """
     Extracts the source of currently scraped page.
         :param page: Selenium Selector, page to export source from
         :param inp: list, inputs from launcher (incl_tables, incl_bullets, output_dir)
     """
 
-    varname = inp[0]
-    filename = inp[1]
+    filename = inp[0]
 
     with open(filename, 'w+', encoding="utf-8") as f:
-        f.write(page.page_source)
+        f.write(browser.page_source)
 
 
-def scan_web_page(page, inp, browser):
+def scan_web_page(browser, page, inp):
     """
     Finds different elements (tables, bullet lists) on page.
         :param page: Selenium Selector, page to search elements in
@@ -116,18 +200,21 @@ def scan_web_page(page, inp, browser):
     incl_texts = inp[2]
     incl_headlines = inp[3]
     incl_links = inp[4]
-    by_xpath = inp[5]
-
-    # Folder for serialized dataframes
-    PICKLE_FOLDER = 'src/pickle_scraped_data'
+    incl_images = inp[5]
+    incl_buttons = inp[6]
+    by_xpath = inp[7]
+    output_folder = inp[8]
 
     # Predefined tags by type
     TABLE_TAG = ['table']
     BULLET_TAGS = ['ul', 'ol']
-    TEXT_TAGS = ['p', 'strong', 'em']  # 'div']
+    TEXT_TAGS = ['p', 'strong', 'em', 'div[normalize-space(text())]', 'span[normalize-space(text())]']
     HEADLINE_TAGS = ['h1', 'h2']
+    IMAGE_TAGS = ['img']
+    BUTTON_TAGS = ['button', 'a[@role="button"]', 'a[contains(@class, "button")]',
+                   'a[contains(@id, "button")]', 'a[@type="button"]', 'a[contains(@class, "btn")]']
 
-    # <a> tags, exceluding links in menu, links as images, mailto links and links with scripts
+    # <a> tags, excluding links in menu, links as images, mailto links and links with scripts
     LINK_TAGS = ["""
                     a[@href
                     and not(contains(@id, "Menu"))  
@@ -145,14 +232,16 @@ def scan_web_page(page, inp, browser):
                        'bullet': BULLET_TAGS,
                        'text': TEXT_TAGS,
                        'headline': HEADLINE_TAGS,
+                       'image': IMAGE_TAGS,
+                       'button': BUTTON_TAGS,
                        'link': LINK_TAGS + ['a']}  # + ['a'] is to identify link tags when using custom XPath
 
     try:
-        shutil.rmtree(PICKLE_FOLDER)
+        shutil.rmtree(output_folder)
     except:
         pass
     finally:
-        os.mkdir(PICKLE_FOLDER)
+        os.mkdir(output_folder)
 
     # Dictionary with elements (XPaths, data)
     final_elements = {}
@@ -161,6 +250,9 @@ def scan_web_page(page, inp, browser):
     innerHTML = browser.execute_script("return document.body.innerHTML")
     tree = lxml.html.fromstring(innerHTML)
 
+    # Url pattern, used to get main page url from current url
+    url_pattern = re.compile('^(((https|http):\/\/|www\.)*[a-zA-Z0-9\.\/\?\:@\-_=#]{2,100}\.[a-zA-Z]{2,6}\/)')
+
     time_start_f = datetime.datetime.now()
 
     def timedelta_format(end, start):
@@ -206,12 +298,21 @@ def scan_web_page(page, inp, browser):
         """
         elems_pos = []
 
+        existing_xpaths = []
+
         for selector, name, xpath, data in zip(selectors, names, xpaths, data):
+
+            # Deduplicate elements by xpath (on some pages the same element is selected twice)
+            if xpath in existing_xpaths:
+                continue
+
             elems_pos.append({'rect': selector.rect,
                               'name': name,
                               'xpath': xpath,
                               'data': data})
 
+            existing_xpaths.append(xpath)
+
         elements_positions = elems_pos
         elements_positions = VarSafe(elements_positions, 'elements_positions', 'elements_positions')
 
@@ -225,7 +326,7 @@ def scan_web_page(page, inp, browser):
             :param xpath: XPath of element
         """
 
-        path = os.path.join(PICKLE_FOLDER, element_name)
+        path = os.path.join(output_folder, element_name)
 
         if 'table' in element_name:
             table_2 = page.xpath(xpath)[0]
@@ -306,11 +407,24 @@ def scan_web_page(page, inp, browser):
                 with open(path + '.pickle', 'wb') as pickle_file:
                     pickle.dump(data, pickle_file)
 
+        elif 'image' in element_name:
+            xpath += '/@src'
+            data = page.xpath(xpath).extract()[0]
+
+            if data:
+                # Handle images with relative path. E.g. images/logo.png
+                if not any([data.startswith(x) for x in ['http', 'www']]):
+                    data = url_pattern.match(browser.current_url).group(1) + data.replace('./', '')
+
+                with open(path + '.pickle', 'wb') as pickle_file:
+                    pickle.dump(data, pickle_file)
         else:
             if 'link' in element_name:
                 xpath += '/@href'
             elif 'element' in element_name:
                 pass
+            elif 'button' in element_name:
+                pass
             else:
                 xpath += '//text()'
 
@@ -343,8 +457,10 @@ def scan_web_page(page, inp, browser):
             prefix = ''
 
         for tag in tags:
-            elements_tree.extend(tree.xpath(f'{prefix}{tag}'))
             elements.extend(browser.find_elements(By.XPATH, f'{prefix}{tag}'))
+            if custom_tag:
+                tag = tag.replace('/body/', '/div/')  # Otherwise, elements_tree will be empty
+            elements_tree.extend(tree.xpath(f'{prefix}{tag}'))
 
         if elements:
             for i, element in enumerate(elements):
@@ -356,6 +472,7 @@ def scan_web_page(page, inp, browser):
                 try:
                     xpath = find_element_xpath(elements_tree, i)
                     serialize_and_append_data(f'{element_name}_{i}', element, xpath)
+
                 except:
                     pass
 
@@ -365,7 +482,6 @@ def scan_web_page(page, inp, browser):
             :param tree: list of elements
             :param i: element's number
         """
-
         xpath = tree[i].getroottree().getpath(tree[i])
         xpath = xpath.split('/')
         xpath[2] = 'body'  # For some reason getpath() generates <div> instead of <body>
@@ -423,8 +539,19 @@ def scan_web_page(page, inp, browser):
     if incl_links:
         find_elements(LINK_TAGS, 'link')
 
+    ##### IMAGES SECTION #####
+    if incl_images:
+        find_elements(IMAGE_TAGS, 'image')
+
+    ##### BUTTONS SECTION #####
+    if incl_buttons:
+        find_elements(BUTTON_TAGS, 'button')
+
     ##### CUSTOM XPATH SECTION #####
     if by_xpath:
+        # With text() at the end will not work
+        by_xpath = by_xpath.removesuffix('/text()').rstrip('/')
+
         custom_tag = [by_xpath]
         custom_tag_splitted = re.split('//|/', by_xpath)  # Split XPath in parts
         last_element_in_xpath = custom_tag_splitted[-1]  # Last element in XPath
@@ -452,7 +579,7 @@ def scan_web_page(page, inp, browser):
     print('[TIME] WHOLE FUNCTION ------>', timedelta_format(datetime.datetime.now(), time_start_f))
 
 
-def wait_until_element_is_located(browser, inp):
+def wait_until_element_is_located(browser, page, inp):
     """
     Waits until certain element is located on page and then clicks on it.
         :param browser: Selenium driver, browser instance
@@ -469,14 +596,16 @@ def wait_until_element_is_located(browser, inp):
         print('Error while locating element', e)
 
 
-def get_current_url(url, inp):
+def get_current_url(browser, page, inp):
     """
     Returns the URL of current opened website
-        :param page: string, URL of opened page from current driver instance
+        :param page: :param browser: driver instance
         :param inp: list, inputs from launcher (filename)
     """
 
     filename = inp[0]
+    url = str(browser.current_url)
+
     try:
         with open(filename, 'w+', encoding="utf-8") as f:
             f.write(url)
@@ -484,7 +613,7 @@ def get_current_url(url, inp):
         print('Error while getting current URL!')
 
 
-def close_browser(browser):
+def close_browser(browser, page, inp):
     """
     Closes browser (removes driver instance).
         :param browser: driver instance
@@ -498,38 +627,44 @@ def close_browser(browser):
             close_browser(browser)
         '''
 
+        # Remove proxy after closing browser instance
+        proxy = {'ip': '', 'port': '', 'username': '', 'password': ''}
+        proxy = VarSafe(proxy, "proxy", "proxy")
+
+        save_variables(kept_variables, 'scr_vars.kpv')
+
     except ConnectionRefusedError:
         pass
     except Exception:
         print('Error while closing the browser!')
 
 
-def scroll_web_page(browser, inp):
+def scroll_web_page(browser, page, inp):
     """
     Scrolls page up / down by n-pixels.
         :param browser: driver instance
-        :param inp: list, inputs from launcher (scroll_to, scroll_by)
+        :param inp: list, inputs from launcher (scroll_to, scroll_by, scroll_max)
     """
 
     scroll_to = inp[0]
     scroll_by = inp[1]
-    #scroll_max = inp[2]
+    scroll_max = inp[2]
 
     script = ''
 
     if scroll_to == 'Down':
-        #if scroll_max:
-            #script = 'window.scrollTo(0, document.body.scrollHeight);var lenOfPage=document.body.scrollHeight;return lenOfPage;'
-        #else:
+        if scroll_max:
+            script = 'window.scrollTo(0, document.body.scrollHeight);var lenOfPage=document.body.scrollHeight;return lenOfPage;'
+        else:
             try:
                 # script = f'window.scrollBy(0, {scroll_by});'       # instant scrolling
                 script = f'window.scrollBy({{top: {scroll_by}, left: 0, behavior: "smooth"}});'  # smooth scrolling
             except:
                 pass
     elif scroll_to == 'Up':
-        #if scroll_max:
-            #script = 'window.scrollTo(0, 0)'
-        #else:
+        if scroll_max:
+            script = 'window.scrollTo(0, 0)'
+        else:
             try:
                 # script = f'window.scrollBy(0, -{scroll_by});'      # instant scrolling
                 script = f'window.scrollBy({{top: -{scroll_by}, left: 0, behavior: "smooth"}});'  # smooth scrolling
@@ -540,16 +675,89 @@ def scroll_web_page(browser, inp):
         browser.execute_script(script)
 
 
-def click_xpath(browser, xpath):
-    browser.find_element_by_xpath(xpath).click()
+def download_images(browser, page, inp):
+    """
+    Downloads images using XPath
+        :param browser: driver instance
+        :param inp: list, inputs from launcher (image xpath, filename)
+    """
+
+    image_xpath = inp[0]
+    filename = inp[1]
+    headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:50.0) Gecko/20100101 Firefox/50.0'}
+
+    # If entered filename contains extension -> drop extension
+    if '.' in filename:
+        filename = filename.split('.')[0]
+
+    images = browser.find_elements(By.XPATH, image_xpath)
+
+    if len(images) == 0:
+        pass
+    elif len(images) == 1:
+        image_url = images[0].get_attribute('data-src')
+
+        if not image_url:
+            image_url = images[0].get_attribute('src')
+
+        image_extension = image_url.split('.')[-1].split('?')[0]  # Drop ?xxx after image extension
+
+        r = requests.get(image_url, headers=headers)
+        with open(f'{filename}.{image_extension}', 'wb') as outfile:
+            outfile.write(r.content)
+    else:
+        # Use provided filename as directory name. Images themselves will be filename_base_0, filename_base_1, filename_base_2 etc.
+        images_directory = filename
+        if not os.path.exists(images_directory):
+            os.mkdir(images_directory)
+
+        filename_base = filename
+
+        for i, image in enumerate(images):
+            try:
+                # Sometimes the image url is stored within data-src tag -> TODO: add new argument to handler with tag?
+                image_url = image.get_attribute('data-src')
+                if not image_url:
+                    image_url = image.get_attribute('src')
+
+                image_extension = image_url.split('.')[-1].split('?')[0]  # Drop ?xxx after image extension
+                filename = f'{filename_base}_{i}.{image_extension}'
+
+                r = requests.get(image_url, headers=headers)
+                with open(f'{images_directory}/{filename}', 'wb') as outfile:
+                    outfile.write(r.content)
+            except:
+                pass
 
 
-def extract_xpath(page, inp):
+def click_xpath(browser, page, inp):
+    xpath = inp[0]
+
+    print('SEARCHING FOR ELEMENT')
+    element = browser.find_element(By.XPATH, xpath)
+
+    if element.is_enabled():
+        print('BUTTON IS ENABLED')
+        element.click()
+    else:
+        print('BUTTON IS NOT ENABLED, ENABLING IT')
+        browser.execute_script("arguments[0].removeAttribute('disabled','disabled')", element)
+        element.click()
+
+def extract_xpath(browser, page, inp):
     """
     write_in_file_mode ... w+, a+
     """
     xpath = inp[0]
     filename = inp[1]  # "extracted_data.txt"
+
+    if not xpath.endswith('/text()') and not '@' in xpath.split('/')[-1]:
+        xpath += '/text()'
+
+    # Extract link from "a" tags
+    if xpath.split('/')[-1] == 'a' or xpath.split('/')[-1] == '/a' or xpath.split('/')[-1].startswith('a['):
+        xpath += '/@href'
+
     try:
         write_in_file_mode = inp[2]
     except:
@@ -558,9 +766,8 @@ def extract_xpath(page, inp):
     data = page.xpath(xpath).extract()
 
     if not data:
-        data = ['EmptyElement']
+        data = ['None']
 
-    # print("DATA",data)
     with open(filename, write_in_file_mode, encoding="utf-8") as f:
         if isinstance(data, list):
             for i, row in enumerate(data):
@@ -575,7 +782,7 @@ def extract_xpath(page, inp):
     # print(data)
 
 
-def extract_multiple_xpaths(page, inp):
+def extract_multiple_xpaths(browser, page, inp):
     print("PAGE", page, "INP", inp)
     result = []
     xpaths = inp[0]
@@ -596,7 +803,7 @@ def extract_multiple_xpaths(page, inp):
         pass
 
 
-def extract_table_xpath(page, inp):
+def extract_table_xpath(browser, page, inp):
     row_xpath = inp[0]
     column_xpath = inp[1]
     filename = inp[2]  # "extracted_data.txt"
@@ -607,7 +814,6 @@ def extract_table_xpath(page, inp):
     ths = page.xpath(row_xpath + '//th')
     headers = []
 
-
     # Try to find headers within <th> tags
     for th_tag in ths:
         headers.append(''.join(th_tag.xpath('.//text()').extract()).replace('\n', '').replace('\t', ''))
@@ -688,37 +894,67 @@ class DocrawlSpider(scrapy.spiders.CrawlSpider):
         #   'CONCURRENT_REQUESTS' : '20',
     }
 
-    # capabilities = DesiredCapabilities.FIREFOX
-    # capabilities["marionette"] = True
-    # browser=webdriver.Firefox(capabilities=capabilities)
-    # browser.set_window_size(1820, 980)
-
     def __init__(self):
         # can be replaced for debugging with browser = webdriver.FireFox()
         # self.browser = webdriver.PhantomJS(executable_path=PHANTOMJS_PATH, service_args=['--ignore-ssl-errors=true'])
-        capabilities = DesiredCapabilities.FIREFOX
-        capabilities["marionette"] = True
-
-        options = FirefoxOptions()
-        window_size_x = 1820
 
         try:
-            bool_scrape_in_browser = load_variable_safe('scr_vars.kpv', 'bool_scrape_in_browser')['in_browser']
+            self.driver_type = load_variable_safe('scr_vars.kpv', 'browser')['driver']
         except Exception as e:
-            print('Error while loading bool_scrape_in_browser!', e)
+            print('Error while loading driver type information: ', e)
+            self.driver_type = 'Firefox'
+
+        try:
+            bool_scrape_in_browser = load_variable_safe('scr_vars.kpv', 'browser')['in_browser']
+        except Exception as e:
+            print('Error while loading headless mode information: ', e)
             bool_scrape_in_browser = True
 
-        if not bool_scrape_in_browser:
-            options.add_argument("--headless")
-
-            # For headless mode different width of window is needed
-            window_size_x = 1450
-
         try:
-            self.browser = webdriver.Firefox(options=options, capabilities=capabilities,
-                                             service=Service(GeckoDriverManager().install()))
-        except:
-            self.browser = webdriver.Firefox(options=options, capabilities=capabilities)
+            proxy_info = load_variable_safe('scr_vars.kpv', 'proxy')
+        except Exception as e:
+            print('Error while loading proxy information: ', e)
+            proxy_info = None
+
+        if self.driver_type == 'Firefox':
+            capabilities = DesiredCapabilities.FIREFOX.copy()
+            self.options = FirefoxOptions()
+            capabilities["marionette"] = True
+
+            sw_options = self._set_proxy(proxy_info)
+
+            if not bool_scrape_in_browser:
+                self.options.add_argument("--headless")
+
+                # For headless mode different width of window is needed
+                window_size_x = 1450
+
+            try:
+                self.browser = webdriver.Firefox(options=self.options, capabilities=capabilities,
+                                                 service=Service(GeckoDriverManager().install()))
+            except Exception as e:
+                print(f'ERROR WHILE CREATING FIREFOX INSTANCE {e}')
+                self.browser = webdriver.Firefox(options=self.options, capabilities=capabilities)
+
+        elif self.driver_type == 'Chrome':
+            capabilities = DesiredCapabilities.CHROME
+            self.options = ChromeOptions()
+
+            sw_options = self._set_proxy(proxy_info)
+
+            if not bool_scrape_in_browser:
+                self.options.add_argument("--headless")
+
+                # For headless mode different width of window is needed
+                window_size_x = 1450
+
+            try:
+                self.browser = webdriver.Chrome(options=self.options, desired_capabilities=capabilities,
+                                                executable_path=ChromeDriverManager().install())
+            except:
+                pass
+
+        window_size_x = 1820
 
         self.browser.set_window_size(window_size_x, 980)
         self.start_requests()
@@ -726,6 +962,49 @@ class DocrawlSpider(scrapy.spiders.CrawlSpider):
     def __del__(self):
         self.browser.quit()
 
+    def _set_proxy(self, proxy_info: dict) -> dict:
+        """
+        Sets proxy before launching browser instance.
+        :param proxy_info: proxy params (ip, port, username, password)
+        """
+
+        # If proxy was not set
+        if proxy_info is None or any([not proxy_info['ip'], not proxy_info['port']]):
+            return None
+
+        proxy_ip = proxy_info['ip']
+        proxy_port = proxy_info['port']
+        proxy_username = proxy_info['username']
+        proxy_password = proxy_info['password']
+
+        if proxy_username and proxy_password:
+            proxy = f'http://{proxy_username}:{proxy_password}@{proxy_ip}:{proxy_port}'
+        else:
+            proxy = f'{proxy_ip}:{proxy_port}'
+
+        # Proxy with authentication
+        if 'http://' in proxy:
+            # selenium-wire proxy settings
+            sw_options = {
+                'proxy': {
+                    'http': proxy,
+                    'https': proxy,
+                    'no_proxy': 'localhost,127.0.0.1'
+                }
+            }
+
+        # Proxy without authentication
+        else:
+            sw_options = None
+            firefox_proxies = Proxy()
+            firefox_proxies.ssl_proxy = proxy
+            firefox_proxies.http_proxy = proxy
+            firefox_proxies.proxy_type = ProxyType.MANUAL
+
+            self.options.proxy = firefox_proxies
+
+        return sw_options
+
     def start_requests(self):
         URLS = ['https://www.forloop.ai']
         FUNCTIONS = [self.parse]
@@ -736,12 +1015,20 @@ class DocrawlSpider(scrapy.spiders.CrawlSpider):
         global spider_functions
         try:
             global browser_pid
-            browser_pid = self.browser.capabilities['moz:processID']
+
+            if self.driver_type == 'Firefox':
+                browser_pid = self.browser.capabilities['moz:processID']
+            elif self.driver_type == 'Chrome':
+                self.browser.service.process
+                browser_pid = psutil.Process(self.browser.service.process.pid)
+                browser_pid = browser_pid.pid
+
             browser_pid = VarSafe(browser_pid, "browser_pid", "browser_pid")
             save_variables(kept_variables, "scr_vars.kpv")
             print(browser_pid)
         except Exception as e:
             print(e)
+
         self.browser.get(response.url)
         global docrawl_core_done
         docrawl_core_done = False
@@ -765,8 +1052,7 @@ class DocrawlSpider(scrapy.spiders.CrawlSpider):
                 time.sleep(1)
                 print("Docrawl core loop")
                 print(spider_functions)
-                # print(docrawl_core_done)
-                # print(spider_requests['loaded'])
+
                 if not spider_requests['loaded']:
                     # print(spider_requests['url'])
                     self.browser.get(spider_requests['url'])
@@ -779,66 +1065,38 @@ class DocrawlSpider(scrapy.spiders.CrawlSpider):
 
                 if spider_functions['done'] == False:
 
-                    # print("AAA",spider_functions)
-                    # try:
-
                     function_str = spider_functions['function']
                     function = eval(function_str)
-                    # print("BBB",function,function_str)
 
-                    print("INPUT", spider_functions['input'])
+                    inp = spider_functions['input']
+                    print('INP FROM DOCRAWL CORE', inp)
 
-                    inp = spider_functions['input']  # .replace("$","'").replace('€','"')
-                    print("INP", inp)
-
-                    if function_str == "click_xpath":
-                        print("CLICK XPATH")
-                        click_xpath(self.browser, inp)
-                    elif function_str == "extract_xpath":
-                        print("EXTRACT XPATH")
-                        extract_xpath(page, inp)
-                    elif function_str == "extract_multiple_xpaths":
-                        print("EXTRACT MULTIPLE XPATH")
-                        extract_multiple_xpaths(page, inp)
-                    elif function_str == "extract_table_xpath":
-                        print("EXTRACT XPATH")
-                        extract_table_xpath(page, inp)
-                    elif function_str == "get_current_url":
-                        print("GET CURRENT URL")
-                        get_current_url(str(self.browser.current_url), inp)
-                    elif function_str == "scan_web_page":
-                        print("SCAN WEB PAGE")
-                        scan_web_page(page, inp, self.browser)
-                    elif function_str == "close_browser":
-                        print("CLOSE BROWSER")
-                        close_browser(self.browser)
-                    elif function_str == "extract_page_source":
-                        print("EXTRACT PAGE SOURCE")
-                        extract_page_source(self.browser, inp)
-                    elif function_str == "wait_until_element_is_located":
-                        print("WAIT UNTIL ELEMENT IS LOCATED")
-                        wait_until_element_is_located(self.browser, inp)
-                    elif function_str == "take_screenshot":
-                        print("TAKE PAGE SCREENSHOT")
-                        take_screenshot(self.browser, inp)
-                    elif function_str == "scroll_web_page":
-                        print("SCROLL WEB PAGE")
-                        scroll_web_page(self.browser, inp)
+                    if function_str in FUNCTIONS.keys():
+                        print(function_str.replace('_', ' ').upper())
+                        FUNCTIONS[function_str](browser=self.browser, page=page, inp=inp)
                     else:
                         function(inp)
 
-                    # except Exception as e:
-                    #    print("Exception occurred:",e)
-                    # print("A")
                     spider_functions['done'] = True
-                    # print("B")
                     spider_functions = VarSafe(spider_functions, "spider_functions", "spider_functions")
-                    # print("C")
                     save_variables(kept_variables, "scr_vars.kpv")
-                    # print("D")
                 page = Selector(text=self.browser.page_source)
                 # save_variables(kept_variables,"scr_vars.kpv")
-                # print("ABCDEFGHIJ",docrawl_core_done)
             except KeyboardInterrupt:
-                # print("BLABLA")
-                break
\ No newline at end of file
+                break
+
+
+FUNCTIONS = {"click_xpath": click_xpath,
+             "extract_xpath": extract_xpath,
+             "extract_multiple_xpaths": extract_multiple_xpaths,
+             "extract_table_xpath": extract_table_xpath,
+             "get_current_url": get_current_url,
+             "scan_web_page": scan_web_page,
+             "close_browser": close_browser,
+             "extract_page_source": extract_page_source,
+             "wait_until_element_is_located": wait_until_element_is_located,
+             "take_screenshot": take_screenshot,
+             "take_png_screenshot": take_png_screenshot,
+             "scroll_web_page": scroll_web_page,
+             "download_images": download_images,
+             }
\ No newline at end of file
